;JT65GEN_FREERUN_SERIAL.ASM    G4JNT  Sept 2011
;JT65 Symbol generator for driving external 4 freq source
; Derived from JT65BCN
;
;Uses more flexible interrupt timing by reprog of TMR0 register 
; test of the New  two-wire Frequency control
; Sent as 32 bit word.  Rx relys on flag and clock count for integrity
; Flag + 24 bit frequency data D3/D2/D1/D0

    LIST P = 16F628A

        INCLUDE "P16F628A.INC"
    __config    0x3F62      ;LV Programming off , OSC HS
    __idlocs    0x802D      ;


    #define     SClk	    PORTA, 0	;Serial Clock signal
    #define     SDat	    PORTA, 1	;Serial Data signal
    #define     SStr	    PORTA, 2	;Serial Strobe

 
    #define	KEYLINE	    PORTB, 2

    #define     TESTLINK    PORTB, 3
    #define     LEDGrn      PORTB, 4
    #define     LEDRed      PORTB, 5

    #define     TXENABLE    PORTB, 6  
    #define     CLOCKTEST   PORTB, 7


    #define     TestModeFlag    Flags, 1
    #define	AllowData	Flags, 2
    #define     CurrData        Flags, 4
    #define     HexFlag         Flags, 6
    #define     AllowSendFlag	Flags, 7


	BOTHMINUTES	=	1	;1 = every minute, 0 = Alternate minutes
    	CWSPEED     	=   	d'50'           ;CW Dot length, ms


			;Values for 19.6608MHz xtal

;	INTDIVIDER	=	d''		;value X, to generate Interupt divisor N (see below)
	JTDIVIDER	=	d'1783'		;  = INTSPERSECOND / 2.69
	INTSPERSECOND	=	d'4800'		;  = Fxtal / 2000  		
	MSDELAY		=	d'245'		;  = Fxtal / 80000 - 1 

	DataFlag	=	0xF8		;Cannot occur in first item of frequency data for any DDS

  errorlevel 1, -207  ;Suppress msgs. and warnings about labels after col 1.
						   
;=================================================================
 org   0x2100
FreqData         ;JT65   based on Fref = 0.1152MHz
 	de	0x37,0x0B,0xA3,0x8E	; JT Sync  1270.46Hz + 23.5kHz
	de	0x01,0x88,0x00,0x00	; 16 * JT65A Interval      
	
JT65MsgData
 include "jt65symb.inc"     ;Jt65 symbols

CWMsg           de      "G4JNT",0

;====================================================================================
	org     0
    nop
    movlw   b'00000000'     ;No interupts until passed thro here once
    movwf   INTCON          ;
    goto    StartUp         ;jump to main code

;------------------ Interrupt routine, get here 4 cycles after TMR overflow --
ints                           
	bcf     INTCON , T0IF   ;		5kHz interrupt
        movwf   Wreg        ;Save W
	swapf   STATUS , W      ; nb.  movwf does not affect any STATUS bits
	movwf   SaveSts         ;Stave STATUS, without affecting it

	bsf	CLOCKTEST
;	movlw	INTDIVIDER
	nop			;nops needed to pad out so total clocks here 
				;  are a multiple of the prescalar value 

;	movwf	TMR0		;TMR0 start recounting from here

		;Interrupt counter goes from X to 256 at Fclock / Prescale.  
		;   Let the overhead = V clocks, then int occurs every (256 - X) * P + V   clocks
		;  X =  256 - (N-V)/P    V must be padded out to be integer multiple of P

;   Below here is not time - critical any more. But must complete in remaining interrupt interval
	btfss	AllowData
	goto	KeepTime
	
	incf    IntCountLo
	btfsc   STATUS, Z
	incf    IntCountHi

        movlw   HIGH(JTDIVIDER)    ;INTPERSECOND / JTDIVIDER = 2.69Hz symbol rate
	subwf   IntCountHi, W
	btfss   STATUS, Z
	goto    KeepTime

        movlw   LOW(JTDIVIDER)
	subwf   IntCountLo, W
	btfss   STATUS, Z
	goto    KeepTime
       ;............             Here at Symbol rateHz
	clrf    IntCountLo
	clrf    IntCountHi

	btfsc   SymbCount, 0    ;Flash Red LED at half symbol rate
        bcf     LEDRed
        btfss   SymbCount, 0
        bsf     LEDRed

        movlw   d'126'
	subwf   SymbCount, W    ;So do the testing for a full message here, 
	btfss   STATUS, Z
	goto    NotLastSymbol

	bcf     AllowData	;Inhibit after final symbol
	goto    KeepTime    ;  and don't do anything else
	;...............
NotLastSymbol

	movf    SymbCount       ;Test for first symbol
	btfsc   STATUS, Z
	call    SyncToStack     ;Load sync pattern into stack first time here

	rlf     Stack0          ;Stack holds sync / message tone pattern
	rlf     Stack1    
	rlf     Stack2
	rlf     Stack3          
	rlf     Stack4
	rlf     Stack5
	rlf     Stack6
	rlf     Stack7
	rlf     Stack8
	rlf     Stack9
	rlf     StackA
	rlf     StackB
	rlf     StackC
	rlf     StackD
	rlf     StackE
	rlf     StackF
	btfss   STATUS, C       ;1 = sync, 0 = character
	goto    JTSymb

	clrf	FreqCache            ;0 passed to routine sends sync tone
	bsf	AllowSendFlag
	goto    SendTone
	;...................
JTSymb                          ;Get next symbol value in message (0-63)
	movf    CharCount, W
	addlw   LOW(JT65MsgData)
	call    GetEE           ;Get value 0 to 63
        addlw   2               ;2* interval Offset from sync (was 3 in JT654)
	movwf	FreqCache
	bsf	AllowSendFlag   ;Pass symbol value to routine,
								; generates frequency data in D0-D3
	incf    CharCount
	;..........
SendTone                        ;Get here with frequency data in D0-5
	incf    SymbCount       ;Increment here, but tested at start allows 
								; the DDS update at last+1 interval, ie. =136

KeepTime
	incf	TimCountLo
	btfsc	STATUS, Z
	incf	TimCountHi

	movlw	HIGH(INTSPERSECOND)
	subwf	TimCountHi, W
	btfss	STATUS, Z
	goto	OutInterrupt

	movlw	LOW(INTSPERSECOND)
	subwf	TimCountLo, W
	btfss	STATUS, Z
	goto	OutInterrupt

	clrf	TimCountHi
	clrf	TimCountLo

	incf	Seconds
	movlw	d'60'
	subwf	Seconds, W
	btfss	STATUS, Z
	goto	OutInterrupt
	clrf	Seconds
	incf	Minutes		;Minutes just roll 0 - 255

OutInterrupt
	bcf	CLOCKTEST                              
	swapf   SaveSts , W     ;Restore STATUS
	movwf   STATUS
	swapf   Wreg            ;Restore W
	swapf   Wreg , W        ;Can't use movf as it  affects Z bit
			
	retfie		;Vaguely 90 clocks or so in here, so OK for 125 division
;=====================================
StartUp 
    clrf    PORTA       ;Special setup to disable comparator on 'F628
    movlw   7
    movwf   CMCON
	
    bsf     STATUS,RP0      ;ram page 1
    movlw   b'00000001'     ;Internal input, Prescalar /4  pull-ups
    movwf   OPTION_REG
    movlw   b'00011000'     ;A3/A4 Frequency bank
    movwf   PORTA           ;
    movlw   b'00001011'     ; B3 TESTLINK 
    movwf   PORTB           ;B6 ext key output, B7 Clock Test output
    bcf     STATUS,RP0      ;ram page 0
    bcf	    SClk	    ;
    bcf	    SStr

    clrf    PORTB
    clrf    PORTA
    clrf    TMR0
    clrf    IntCountHi
    clrf    IntCountLo
    clrf    Flags
    clrf    CharCount
    clrf    SymbCount
    clrf    Seconds	  
    clrf    Minutes
			;Launch straight in to sending, assuming unit has been reset at the minute point
    bcf     LEDGrn
    movlw   b'10000000'     ;Enable global int, not timer overflow int yet
    movwf   INTCON
    bsf     INTCON, T0IE
;...........................
MainLoop
    btfss   TESTLINK
    call    TestMode

  if BOTHMINUTES == 0

    btfsc   Minutes, 0      ;Test for the odd minute
    goto    CarryOn

  endif

    movf    Seconds, W		;At start, this happens even before the interrupt has updated the time
    sublw   d'00'
    btfss   STATUS, Z
    goto    CarryOn 
                      ;Now at even minute : 00

    bsf     TXENABLE    ;Enable Tx up to 1s earlier to allow for relays
    call    LongDelay	;Start at about 1s interval
    bsf     KEYLINE
    bcf     LEDGrn

    movlw   0
    call    SendFreq

    clrf    IntCountLo
    clrf    IntCountHi
    clrf    CharCount
    clrf    SymbCount    ;Re-Initialise registers

    bsf     AllowData	 ;Starts JT65 sequence

JT65Loop

	btfss	AllowSendFlag
	goto	NoSend
	movf	FreqCache, W
	call	SendFreq	;Has to be done outside interrupt routine
	bcf	AllowSendFlag	; as and interrupt overrun would occur
NoSend
    btfsc   AllowData	 ;Sit in here waiting for completion   
    goto    JT65Loop

    movlw   0
    call    SendFreq

    bcf     KEYLINE
    call    LongDelay

    movlw   LOW(CWMsg)
    movwf   CWMsgCnt
    call    SendCWMsg

    bcf     TXENABLE
    bsf     LEDGrn
    bcf	    LEDRed
CarryOn
    goto    MainLoop
;.............................

;===========================================
TestMode                    ;TESTLINK Input grounded
    bsf     LEDRed
    bcf     LEDGrn
    movlw   0
    call    SendFreq	
    bsf     TXENABLE
    bsf     KEYLINE
TestModeLoop
    btfss   TESTLINK
    goto    TestModeLoop

    bcf     LEDRed
    bcf     TXENABLE
    bcf     KEYLINE
    return
;--------------------------------
;----------------------
Multiply32  ;32 bit * 32 bit multiplication, more efficient method
    Clrf    Q0        ;Input A0/1/2/3 and B0/1/2/3, output Q0/2/3/4/5/6/7/8
    clrf    Q1        ;A destroyed,  B kept
    clrf    Q2        
    clrf    Q3       
    clrf    Q4
    clrf    Q5
    clrf    Q6
    clrf    Q7

    movlw   d'32'       ;A and B 24 bit
    movwf   Counter
MultLoop            
    bcf     STATUS, C
    rrf     A3
    rrf     A2          ;Working bit into carry
    rrf     A1
    rrf     A0          ;Effectivley adds shifted version of B
    btfss   STATUS, C   ; into Acc depending on working bit of A
    goto    NoMult      ; by shifting the accumulator instead of B itself

    movf    B0, W       ;get shifted B0 value
    addwf   Q4        ; and add into accumulator if 1 in 
    btfss   STATUS, C   ; appropriate bit in Avalue
    goto    MulAdd1
    incf    Q5
    btfss   STATUS, Z
    goto    MulAdd1     
    incf    Q6
    btfss   STATUS, Z
    goto    MulAdd1
    movlw   1           ;Need more complex addition routine for last Acc as
    addwf   Q7        ; final Carry has to be preserved for the shift
MulAdd1
    movf    B1, W
    addwf   Q5
    btfss   STATUS, C
    goto    MulAdd2
    incf    Q6
    btfss   STATUS, Z
    goto    MulAdd2
    movlw   1
    addwf   Q7

MulAdd2
    movf    B2, W
    addwf   Q6
    btfss   STATUS, C
    goto    MulAdd3
    movlw   1
    addwf   Q7
MulAdd3
    movf    B3, W
    addwf   Q7

NoMult                  ;Need to rotate in carry from final addition.  C is
    rrf     Q7        ; already zero if we've jumped in here.
    rrf     Q6
    rrf     Q5        
    rrf     Q4
    rrf     Q3
    rrf     Q2
    rrf     Q1
    rrf     Q0
    decfsz  Counter
    goto    MultLoop

    return              ;Q7/7/6/5/4/3/2/1 = A * B     
;------------------------------------------------------------------
SyncToStack
	movlw   0x98
	movwf   StackF
	movlw   0xFD
	movwf   StackE
	movlw   0x45
	movwf   StackD
	movlw   0x91
	movwf   StackC
	movlw   0xCF
	movwf   StackB
	movlw   0x6F
	movwf   StackA
	movlw   0x1A
	movwf   Stack9
	movlw   0xB3
	movwf   Stack8
	movlw   0x54
	movwf   Stack7
	movlw   0x81
	movwf   Stack6
	movlw   0x80
	movwf   Stack5
	movlw   0xD2
	movwf   Stack4
	movlw   0xD5
	movwf   Stack3
	movlw   0x32
	movwf   Stack2
	movlw   0x43
	movwf   Stack1
	movlw   0xFC
	movwf   Stack0

	return
;---------------------------------------
GetEE         ;different from ;F84   EE registers all in bank 1 now 
	bsf     STATUS,RP0      ;ram page 1
	movwf   EEADR
	bsf     EECON1 , RD
	movf    EEDATA , W
	bcf     STATUS , RP0    ;ram page 0
	return
;-----------------------
LongDelay           ;Approx 1 second delay
	movlw   d'250'
	movwf   DelCount2
LongDelLoop
	call    Delay1ms
	call    Delay1ms
	call    Delay1ms
	call    Delay1ms
	decfsz  DelCount2
	goto    LongDelLoop
	return
;-------------------------
Delay1ms        ;20N+5
	movlw   MSDELAY
	movwf   DelCount
Del1msLoop
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	nop
	decfsz  DelCount
	goto    Del1msLoop
	return
;-------------------------
SendFreq
	movwf	B0		;TOne increment Value from 0 to 65 
	clrf	B1
	clrf	B2
	clrf	B3

		;Read the user switches on PORTA3/A4 for A/B/C selection and multiply  the tone offset  accordingly
	
	bcf	STATUS, C
	btfsc	PORTA, 3	;00 = JT65A   01 or 10 = JT65B   11 = JT65C
	rlf	B0	

	bcf	STATUS, C
	btfss	PORTA, 4
	goto	NotMSSw
	rlf	B0	
	rlf	B1		;B0/1 could range from 0 to d'260' for JT65C

NotMSSw
	movlw	LOW(FreqData + 4)	;Point to EE location storing the increment * 256 for JT65
	call	GetEE
	movwf	A3	
	movlw	LOW(FreqData + 5)
	call	GetEE
	movwf	A2
	movlw	LOW(FreqData + 6)
	call	GetEE
	movwf	A1
	movlw	LOW(FreqData + 7)
	call	GetEE
	movwf	A0

	call	Multiply32	;Gives tone offset * 256 in Q0-7

	movlw	LOW(FreqData)	;Point to EE location storing the tone reference
	call	GetEE
	movwf	A3	
	movlw	LOW(FreqData + 1)
	call	GetEE
	movwf	A2
	movlw	LOW(FreqData + 2)
	call	GetEE
	movwf	A1
	movlw	LOW(FreqData + 3)
	call	GetEE
	movwf	A0


		
	movf    Q1, W		;True total offset in Q1 - Q3 (Q is divided by 256)
	addwf   A0
	btfss   STATUS, C
	goto    TonAdd1
	incf    A1
	btfss   STATUS, Z
	goto    TonAdd1
	incf    A2
	btfss   STATUS, Z
	goto    TonAdd1
	incf    A3
		
TonAdd1
	movf    Q2, W
	addwf   A1
	btfss   STATUS, C
	goto    TonAdd2
	incf    A2
	btfss   STATUS, Z
	goto    TonAdd2
	incf    A3
	
TonAdd2
	movf    Q3, W
	addwf   A2
	btfss   STATUS, C
	goto    TonAdd3
	incf    A3
TonAdd3
	movf    Q4, W
	addwf   A3		;A0-3 Contains 32 frequency value to send to DDS
			;  But only 24 bits are used, with a header
			; So for neatness, transfer to D0-3 as done in the JT65 version

ToneMultDone
	movf	A3, W
	movwf	D2
	movf	A2, W
	movwf	D1
	movf	A1, W
	movwf	D0
	movlw	DataFlag	;Alignment check flag
	movwf	D3

		;~~~~~~~~   Now send the data ~~~~~~~~~~~~
	movlw	d'32'
	movwf	Counter
SendSerLoop
	bcf	SClk		;Data changes on negative edge
	rlf	D0
	rlf	D1
	rlf	D2		;Send MSB first, starting with the Flag info	
	rlf	D3		
	btfsc	STATUS, C
	bsf	SDat	
	btfss	STATUS, C
	bcf	SDat
	call	Delay1ms
	bsf	SClk
	call	Delay1ms
	decfsz	Counter
	goto	SendSerLoop	;32 low to high transitions of SClk

	bcf	SClk		;
	call	Delay1ms
	bsf	SStr		;Str high after clock goes low is the reference
	call	Delay1ms	;Pulse the Strobe line
	bcf	SStr

	return

;-------------------------------
PushStack
	movf    StackE , W
	movwf   StackF
	movf    StackD , W  
	movwf   StackE
	movf    StackC , W
	movwf   StackD
	movf    StackB , W
	movwf   StackC
	movf    StackA , W
	movwf   StackB
	movf    Stack9 , W
	movwf   StackA
	movf    Stack8 , W
	movwf   Stack9
	movf    Stack7 , W
	movwf   Stack8
	movf    Stack6 , W
	movwf   Stack7 
	movf    Stack5 , W
	movwf   Stack6
	movf    Stack4 , W
	movwf   Stack5
	movf    Stack3 , W
	movwf   Stack4
	movf    Stack2 , W
	movwf   Stack3
	movf    Stack1 , W
	movwf   Stack2
	movf    Stack0 , W
	movwf   Stack1

	return
;-------------------------------
SendCWMsg           ;Enter with EE address programmed into CWMsgCnt
	movf    CWMsgCnt, W
    call    GetEE       ;Test for terminator at start = no CW message
	addlw   0
	btfsc   STATUS, Z
	return
    bcf     KEYLINE
	call    CWDelay
	call    CWDelay
	call    CWDelay
	call    CWDelay

CWLoop
	movf    CWMsgCnt, W
	call    GetEE
	addlw   0
	btfsc   STATUS, Z
	goto    DoneCWMsg

	call    SendCWChar
	incf    CWMsgCnt
	goto    CWLoop
DoneCWMsg
	call    CWDelay
	call    CWDelay
	call    CWDelay
	call    CWDelay
	return
;--------------------------
SendCWChar      ;Sends a single CW character
	addlw   0xE0     ;Subtract 32 (add 256 - 32) to point to table
	call    CWTable
	movwf   CWChar      ;Doesn't affect STATUS so...
	movf    CWChar      ;  check for invalid returned code here
	btfsc   STATUS , Z      
	goto    NoChar

	movlw   0x07    
	andwf   CWChar , W      ;Extract element count into W
	movwf   CWCharCnt     
    movlw   6       
	subwf   CWCharCnt , W     ;W = Elcount - 6.  If +ve/0 then C=1, > 5 els
	btfss   STATUS , C
	goto    Send5
	
	rrf     CWCharCnt   ;Puts LSB into C as 110 = .  111 = -
	call    SendEle     ;First  . or -  for 6 element characters
	movlw   5
	movwf   CWCharCnt     ;Force count to 5 for rest of the elements
Send5       
	rlf     CWChar      ;Element to be sent into C.
	call    SendEle     ;   count already in Counter
	decfsz  CWCharCnt   
	goto    Send5
NoChar          ;Inter word gap, and invalid codes.
	bcf     LEDRed
	call    CWDelay
	call    CWDelay
	return
;-------------------
SendEle         ;Data in STATUS , C;  0 = dit  1 = dah
	rrf     Stack0      ;Save C bit temporarily on stack
    bsf     KEYLINE
	bsf     LEDRed
	rlf     Stack0      ;Recovery C bit again, to get dot/dash 
	btfss   STATUS , C      ;test for dot or dash
	goto    SendDot
	call    CWDelay
	call    CWDelay
SendDot
	call    CWDelay
    bcf     KEYLINE
	bcf     LEDRed
    call    CWDelay     ;Inter element gap
	return
;-------------------
CWDelay       
	movlw   CWSPEED
	movwf   DelCount2
CWDelLoop
	call    Delay1ms      
	decfsz  DelCount2
	goto    CWDelLoop
	return
;--------------------------
UpCase
    return

;-------------------------


;============================================


org 0x300

CWTable       ;CW Characters in compressed form. 
	movwf   Temp
	movlw   HIGH($)
    movwf   PCLATH    
	movf    Temp , W
	addwf   PCL
	
	retlw   b'00000000'  ; [sp] 00
	retlw   b'00000000'  ;  !   00
	retlw   b'10010110'  ;  "   96
	retlw   b'00010101'  ;  #   15
	retlw   b'00000000'  ;  $   00
	retlw   b'00000000'  ;  %   00
	retlw   b'00000000'  ;  &   00
	retlw   b'11110110'  ;  '   F6
	retlw   b'10110101'  ;  (   B5
	retlw   b'01101111'  ;  )   6F
	retlw   b'01000101'  ;  *   45
	retlw   b'00000000'  ;  +   00
	retlw   b'10011111'  ;  ,   9F     Coding for CW characters of up to
	retlw   b'00001111'  ;  -   0F      6 elements.  Stored as :
	retlw   b'10101110'  ;  .   AE     Bits 7 to 3 make first five CW elements
	retlw   b'10010101'  ;  /   95     Reading  Left to Right 
	retlw   b'11111101'  ;  0   FD     Coded as 0 = dit   1 = dah
	retlw   b'01111101'  ;  1   7D     Bts 2-0 are number of elements except for 
	retlw   b'00111101'  ;  2   3D     110 > 1st element is a dit  
	retlw   b'00011101'  ;  3   1D     111 > 1st element is a dah       
	retlw   b'00001101'  ;  4   0D
	retlw   b'00000101'  ;  5   05
	retlw   b'10000101'  ;  6   85
	retlw   b'11000101'  ;  7   C5
	retlw   b'11100101'  ;  8   E5
	retlw   b'11110101'  ;  9   F5
	retlw   b'00000000'  ;  :   00
	retlw   b'00000000'  ;  ;   00
	retlw   b'00000000'  ;  <   00
	retlw   b'10001101'  ;  =   8D
	retlw   b'00000000'  ;  >   00
	retlw   b'01100110'  ;  ?   66
	retlw   b'11010110'  ;  @   D6
	retlw   b'01000010'  ;  A   42
	retlw   b'10000100'  ;  B   84
	retlw   b'10100100'  ;  C   A4
	retlw   b'10000011'  ;  D   83
	retlw   b'00000001'  ;  E   01
	retlw   b'00100100'  ;  F   24
	retlw   b'11000011'  ;  G   C3
	retlw   b'00000100'  ;  H   04
	retlw   b'00000010'  ;  I   02
	retlw   b'01110100'  ;  J   74
	retlw   b'10100011'  ;  K   A3
	retlw   b'01000100'  ;  L   44
	retlw   b'11000010'  ;  M   C2
	retlw   b'10000010'  ;  N   82
	retlw   b'11100011'  ;  O   E3
	retlw   b'01100100'  ;  P   64
	retlw   b'11010100'  ;  Q   D4
	retlw   b'01000011'  ;  R   43
	retlw   b'00000011'  ;  S   03
	retlw   b'10000001'  ;  T   81
	retlw   b'00100011'  ;  U   23
	retlw   b'00010100'  ;  V   14
	retlw   b'01100011'  ;  W   63
	retlw   b'10010100'  ;  X   94
	retlw   b'10110100'  ;  Y   B4
	retlw   b'11000100'  ;  Z   C4


;================================
 
;--------------------------------
    cblock 0x20
        Temp
        DelCount, DelCount2
        IntCountLo,  IntCountHi
	TimCountLo, TimCountHi
	FreqCache
        Counter
        SymbCount       ;Counts 0 to 126 for each JT65 symbol
	CharCount
        JT65Data         ;Used in getting the freq data from EE
        PosData         ;Position of the 2 data bits in Byte from EE
        SoftCount       ;Used in amplitude ramping
        Amplitude
        Flags
        Wreg
        SaveSts
        Seconds, Minutes
	BCDHi, BCDLo
	A0, A1, A2, A3
	B0, B1, B2, B3
	Q0, Q1, Q2, Q3, Q4, Q5, Q6, Q7
        ICount
        CommaCount
        CWChar, CWMsgCnt, CWCharCnt
        RxChar
        TonePtr
        SerCounter      
        AddrHold
        MsgCount
	D0, D1, D2, D3	;Frequency Data sent on serial interface
        Stack0, Stack1, Stack2, Stack3, Stack4, Stack5, Stack6, Stack7 
	Stack8, Stack9, StackA, StackB, StackC, StackD, StackE, StackF
    endc   
;==========================================
		end
