;WSPRGEN_FREERUN_SERIAL.ASM    G4JNT  Sept 2011
;WSPR Symbol generator for driving external 4 freq source
; Derived from WSPRBCN
;
;Uses more flexible interrupt timing by reprog of TMR0 register 
; test of the New  two-wire Frequency control
; Sent as 32 bit word.  Rx relys on flag and clock count for integrity
; Flag + 24 bit frequency data D3/D2/D1/D0

    LIST P = 16F628A

        INCLUDE "P16F628A.INC"
    __config    0x3F62      ;LV Programming off , OSC HS
    __idlocs    0x8027      ;


    #define     SClk	    PORTA, 0	;Serial Clock signal
    #define     SDat	    PORTA, 1	;Serial Data signal
    #define     SStr	    PORTA, 2	;Serial Strobe

 
    #define	KEYLINE	    PORTB, 2

    #define     TESTLINK    PORTB, 3
    #define     LEDGrn      PORTB, 4
    #define     LEDRed      PORTB, 5

    #define     TXENABLE    PORTB, 6  
    #define     CLOCKTEST   PORTB, 7


    #define     TestModeFlag    Flags, 1
    #define	AllowData	Flags, 2
    #define     CurrData        Flags, 4
    #define     HexFlag         Flags, 6
    #define     AllowSendFlag	Flags, 7


	BOTHMINUTES	=	1	;1 = every minute, 0 = Alternate minutes
    	CWSPEED     	=   	d'50'           ;CW Dot length, ms


			;Values for 19.6608MHz xtal

;	INTDIVIDER	=	d''		;value X, to generate Interupt divisor N (see below)
	JTDIVIDER	=	d'3277'		;  = INTSPERSECOND / Symbol rate
	INTSPERSECOND	=	d'4800'		;  = Fxtal / 2000  		
	MSDELAY		=	d'245'		;  = Fxtal / 80000 - 1 

	DataFlag	=	0xF8		;Cannot occur in first item of frequency data for any DDS

  errorlevel 1, -207  ;Suppress msgs. and warnings about labels after col 1.
						   
;=================================================================
 org   0x2100
FreqData           ;WSPR Data  for Fclock = 115.2kHz

    de  0x37, 0x8C, 0xF8, 0xE4    ;Tone 0   24997.803Hz
    de  0x37, 0x8D, 0xCE, 0x39    ;Tone 1   24999.268Hz
    de  0x37, 0x8E, 0xA3, 0x8F    ;Tone 2   25000.732Hz
    de  0x37, 0x8F, 0x78, 0xE4    ;Tone 3   25002.197Hz



	
WSPRMsgData
 include "WSPTONEs.INC"     ;Left Justified WSPR tones, 4 per byte

CWMsg           de      "G4JNT",0

;====================================================================================
	org     0
    nop
    movlw   b'00000000'     ;No interupts until passed thro here once
    movwf   INTCON          ;
    goto    StartUp         ;jump to main code

;------------------ Interrupt routine, get here 4 cycles after TMR overflow --
ints                           
	bcf     INTCON , T0IF   ;		5kHz interrupt
        movwf   Wreg        ;Save W
	swapf   STATUS , W      ; nb.  movwf does not affect any STATUS bits
	movwf   SaveSts         ;Stave STATUS, without affecting it

	bsf	CLOCKTEST
;	movlw	INTDIVIDER
	nop			;nops needed to pad out so total clocks here 
				;  are a multiple of the prescalar value 

;	movwf	TMR0		;TMR0 start recounting from here

		;Interrupt counter goes from X to 256 at Fclock / Prescale.  
		;   Let the overhead = V clocks, then int occurs every (256 - X) * P + V   clocks
		;  X =  256 - (N-V)/P    V must be padded out to be integer multiple of P

;   Below here is not time - critical any more. But must complete in remaining interrupt interval
	btfss	AllowData
	goto	KeepTime
	
	incf    IntCountLo
	btfsc   STATUS, Z
	incf    IntCountHi

        movlw   HIGH(JTDIVIDER)    ;INTPERSECOND / JTDIVIDER = 4.375Hz symbol rate
	subwf   IntCountHi, W
	btfss   STATUS, Z
	goto    KeepTime

        movlw   LOW(JTDIVIDER)
	subwf   IntCountLo, W
	btfss   STATUS, Z
	goto    KeepTime
       ;............             Here at Symbol rateHz
	clrf    IntCountLo
	clrf    IntCountHi

	btfsc   SymbCount, 0    ;Flash Red LED at half symbol rate
        bcf     LEDRed
        btfss   SymbCount, 0
        bsf     LEDRed

        movlw   d'162'
	subwf   SymbCount, W    ;So do the testing for a full message here, 
	btfss   STATUS, Z
	goto    NotLastSymbol

	bcf     AllowData	;Inhibit after final symbol
	goto    KeepTime    ;  and don't do anything else
	;...............
NotLastSymbol
        rrf     SymbCount, W
        movwf   WSPRData        ;WSPRData used as a temporary variable here
        rrf     WSPRData, W
        andlw   0x3F            ;W = SymbCount \ 4, Posn. of this data in EE
        addlw   LOW(WSPRMsgData)
        call    GetEE       ;each data byte is called 4 times, for each pair
        movwf   WSPRData        ;WSPRData contains 4 items of freq data

        movf    SymbCount, W
        andlw   0x03            ;Get position in the item (LH = first)
        movwf   PosData
        incf    PosData         ;want 1,2,3 or 4 passes through alignment loop
PosDataLoop
        rlf     WSPRData        ;Data items are left justified in each byte
        rlf     WSPRData        ;shift left 2, 4, or 6 times though C
        decfsz  PosData
        goto    PosDataLoop

        rlf     WSPRData, W     ;One final shift to position at LSBs
        andlw   0x03            ; total of 3,5,7, or 9 left shifts

	movwf	FreqCache
	bsf	AllowSendFlag

	incf    SymbCount       ;Increment here, but tested at start allows 
				; the freq update at last+1 interval, ie. 207
KeepTime
	incf	TimCountLo
	btfsc	STATUS, Z
	incf	TimCountHi

	movlw	HIGH(INTSPERSECOND)
	subwf	TimCountHi, W
	btfss	STATUS, Z
	goto	OutInterrupt

	movlw	LOW(INTSPERSECOND)
	subwf	TimCountLo, W
	btfss	STATUS, Z
	goto	OutInterrupt

	clrf	TimCountHi
	clrf	TimCountLo

	incf	Seconds
	movlw	d'60'
	subwf	Seconds, W
	btfss	STATUS, Z
	goto	OutInterrupt
	clrf	Seconds
	incf	Minutes		;Minutes just roll 0 - 255

OutInterrupt
	bcf	CLOCKTEST                              
	swapf   SaveSts , W     ;Restore STATUS
	movwf   STATUS
	swapf   Wreg            ;Restore W
	swapf   Wreg , W        ;Can't use movf as it  affects Z bit
			
	retfie		;Vaguely 90 clocks or so in here, so OK for 125 division
;=====================================
StartUp 
    clrf    PORTA       ;Special setup to disable comparator on 'F628
    movlw   7
    movwf   CMCON
	
    bsf     STATUS,RP0      ;ram page 1
    movlw   b'00000001'     ;Internal input, Prescalar /4  pull-ups
    movwf   OPTION_REG
    movlw   b'00011000'     ;A3/A4 Frequency bank
    movwf   PORTA           ;
    movlw   b'00001011'     ; B3 TESTLINK 
    movwf   PORTB           ;B6 ext key output, B7 Clock Test output
    bcf     STATUS,RP0      ;ram page 0
    bcf	    SClk	    ;
    bcf	    SStr

    clrf    PORTB
    clrf    PORTA
    clrf    TMR0
    clrf    IntCountHi
    clrf    IntCountLo
    clrf    SymbCount
    clrf    Flags
    clrf    Seconds	  
    clrf    Minutes
			;Launch straight in to sending, assuming unit has been reset at the minute point
    bcf     LEDGrn
    movlw   b'10000000'     ;Enable global int, not timer overflow int yet
    movwf   INTCON
    bsf     INTCON, T0IE
;...........................
MainLoop
    btfss   TESTLINK
    call    TestMode


    btfsc   Minutes, 0      ;Test for the odd minute
    goto    CarryOn

    movf    Seconds, W		;At start, this happens even before the interrupt has updated the time
    sublw   d'00'
    btfss   STATUS, Z
    goto    CarryOn 
                      ;Now at even minute : 00

    bsf     TXENABLE    ;Enable Tx up to 1s earlier to allow for relays
    call    LongDelay	;Start at about 1s interval
    bsf     KEYLINE
    bcf     LEDGrn

    movlw   0
    call    SendFreq

    clrf    IntCountLo
    clrf    IntCountHi
    clrf    SymbCount    ;Re-Initialise registers

    bsf     AllowData	 ;Starts WSPR sequence

WSPRLoop

	btfss	AllowSendFlag
	goto	NoSend
	movf	FreqCache, W
	call	SendFreq	;Has to be done outside interrupt routine
	bcf	AllowSendFlag	; as and interrupt overrun would occur
NoSend
    btfsc   AllowData	 ;Sit in here waiting for completion   
    goto    WSPRLoop

    movlw   0
    call    SendFreq

    bcf     KEYLINE
    call    LongDelay

    movlw   LOW(CWMsg)
    movwf   CWMsgCnt
    call    SendCWMsg

    bcf     TXENABLE
    bsf     LEDGrn
    bcf	    LEDRed
CarryOn
    goto    MainLoop
;.............................

;===========================================
TestMode                    ;TESTLINK Input grounded
    bsf     LEDRed
    bcf     LEDGrn
    movlw   0
    call    SendFreq	
    bsf     TXENABLE
    bsf     KEYLINE
TestModeLoop
    btfss   TESTLINK
    goto    TestModeLoop

    bcf     LEDRed
    bcf     TXENABLE
    bcf     KEYLINE
    return
;--------------------------------

GetEE         ;different from ;F84   EE registers all in bank 1 now 
	bsf     STATUS,RP0      ;ram page 1
	movwf   EEADR
	bsf     EECON1 , RD
	movf    EEDATA , W
	bcf     STATUS , RP0    ;ram page 0
	return
;-----------------------
LongDelay           ;Approx 1 second delay
	movlw   d'250'
	movwf   DelCount2
LongDelLoop
	call    Delay1ms
	call    Delay1ms
	call    Delay1ms
	call    Delay1ms
	decfsz  DelCount2
	goto    LongDelLoop
	return
;-------------------------
Delay1ms        ;20N+5
	movlw   MSDELAY
	movwf   DelCount
Del1msLoop
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	nop
	decfsz  DelCount
	goto    Del1msLoop
	return
;-------------------------
SendFreq
	movwf	Temp		;Save two bit frequency code
	rlf	Temp
	rlf	Temp, W
	andlw	0x0C		;Multiply by 4
	addlw	LOW(FreqData)
	movwf	Temp		;Temp is pointer to frequencty data for this symbol

	movf	Temp, W
	call	GetEE
	movwf	D2		;Stored MSB first

	movf	Temp, W
	addlw	1
	call	GetEE
	movwf	D1

	movf	Temp, W
	addlw	2
	call	GetEE
	movwf	D0

	movlw	DataFlag	;Alignment check flag
	movwf	D3


	movlw	d'32'
	movwf	Counter
SendSerLoop
	bcf	SClk		;Data changes on negative edge
	rlf	D0
	rlf	D1
	rlf	D2		;Send MSB first, starting with the Flag info	
	rlf	D3		
	btfsc	STATUS, C
	bsf	SDat	
	btfss	STATUS, C
	bcf	SDat
	call	Delay1ms
	bsf	SClk
	call	Delay1ms
	decfsz	Counter
	goto	SendSerLoop	;32 low to high transitions of SClk

	bcf	SClk		;
	call	Delay1ms
	bsf	SStr		;Str high after clock goes low is the reference
	call	Delay1ms	;Pulse the Strobe line
	bcf	SStr

	return

;-------------------------------
PushStack
	movf    StackE , W
	movwf   StackF
	movf    StackD , W  
	movwf   StackE
	movf    StackC , W
	movwf   StackD
	movf    StackB , W
	movwf   StackC
	movf    StackA , W
	movwf   StackB
	movf    Stack9 , W
	movwf   StackA
	movf    Stack8 , W
	movwf   Stack9
	movf    Stack7 , W
	movwf   Stack8
	movf    Stack6 , W
	movwf   Stack7 
	movf    Stack5 , W
	movwf   Stack6
	movf    Stack4 , W
	movwf   Stack5
	movf    Stack3 , W
	movwf   Stack4
	movf    Stack2 , W
	movwf   Stack3
	movf    Stack1 , W
	movwf   Stack2
	movf    Stack0 , W
	movwf   Stack1

	return
;-------------------------------
SendCWMsg           ;Enter with EE address programmed into CWMsgCnt
	movf    CWMsgCnt, W
    call    GetEE       ;Test for terminator at start = no CW message
	addlw   0
	btfsc   STATUS, Z
	return
    bcf     KEYLINE
	call    CWDelay
	call    CWDelay
	call    CWDelay
	call    CWDelay

CWLoop
	movf    CWMsgCnt, W
	call    GetEE
	addlw   0
	btfsc   STATUS, Z
	goto    DoneCWMsg

	call    SendCWChar
	incf    CWMsgCnt
	goto    CWLoop
DoneCWMsg
	call    CWDelay
	call    CWDelay
	call    CWDelay
	call    CWDelay
	return
;--------------------------
SendCWChar      ;Sends a single CW character
	addlw   0xE0     ;Subtract 32 (add 256 - 32) to point to table
	call    CWTable
	movwf   CWChar      ;Doesn't affect STATUS so...
	movf    CWChar      ;  check for invalid returned code here
	btfsc   STATUS , Z      
	goto    NoChar

	movlw   0x07    
	andwf   CWChar , W      ;Extract element count into W
	movwf   CWCharCnt     
    movlw   6       
	subwf   CWCharCnt , W     ;W = Elcount - 6.  If +ve/0 then C=1, > 5 els
	btfss   STATUS , C
	goto    Send5
	
	rrf     CWCharCnt   ;Puts LSB into C as 110 = .  111 = -
	call    SendEle     ;First  . or -  for 6 element characters
	movlw   5
	movwf   CWCharCnt     ;Force count to 5 for rest of the elements
Send5       
	rlf     CWChar      ;Element to be sent into C.
	call    SendEle     ;   count already in Counter
	decfsz  CWCharCnt   
	goto    Send5
NoChar          ;Inter word gap, and invalid codes.
	bcf     LEDRed
	call    CWDelay
	call    CWDelay
	return
;-------------------
SendEle         ;Data in STATUS , C;  0 = dit  1 = dah
	rrf     Stack0      ;Save C bit temporarily on stack
    bsf     KEYLINE
	bsf     LEDRed
	rlf     Stack0      ;Recovery C bit again, to get dot/dash 
	btfss   STATUS , C      ;test for dot or dash
	goto    SendDot
	call    CWDelay
	call    CWDelay
SendDot
	call    CWDelay
    bcf     KEYLINE
	bcf     LEDRed
    call    CWDelay     ;Inter element gap
	return
;-------------------
CWDelay       
	movlw   CWSPEED
	movwf   DelCount2
CWDelLoop
	call    Delay1ms      
	decfsz  DelCount2
	goto    CWDelLoop
	return
;--------------------------
UpCase
    return

;-------------------------


;============================================


org 0x300

CWTable       ;CW Characters in compressed form. 
	movwf   Temp
	movlw   HIGH($)
    movwf   PCLATH    
	movf    Temp , W
	addwf   PCL
	
	retlw   b'00000000'  ; [sp] 00
	retlw   b'00000000'  ;  !   00
	retlw   b'10010110'  ;  "   96
	retlw   b'00010101'  ;  #   15
	retlw   b'00000000'  ;  $   00
	retlw   b'00000000'  ;  %   00
	retlw   b'00000000'  ;  &   00
	retlw   b'11110110'  ;  '   F6
	retlw   b'10110101'  ;  (   B5
	retlw   b'01101111'  ;  )   6F
	retlw   b'01000101'  ;  *   45
	retlw   b'00000000'  ;  +   00
	retlw   b'10011111'  ;  ,   9F     Coding for CW characters of up to
	retlw   b'00001111'  ;  -   0F      6 elements.  Stored as :
	retlw   b'10101110'  ;  .   AE     Bits 7 to 3 make first five CW elements
	retlw   b'10010101'  ;  /   95     Reading  Left to Right 
	retlw   b'11111101'  ;  0   FD     Coded as 0 = dit   1 = dah
	retlw   b'01111101'  ;  1   7D     Bts 2-0 are number of elements except for 
	retlw   b'00111101'  ;  2   3D     110 > 1st element is a dit  
	retlw   b'00011101'  ;  3   1D     111 > 1st element is a dah       
	retlw   b'00001101'  ;  4   0D
	retlw   b'00000101'  ;  5   05
	retlw   b'10000101'  ;  6   85
	retlw   b'11000101'  ;  7   C5
	retlw   b'11100101'  ;  8   E5
	retlw   b'11110101'  ;  9   F5
	retlw   b'00000000'  ;  :   00
	retlw   b'00000000'  ;  ;   00
	retlw   b'00000000'  ;  <   00
	retlw   b'10001101'  ;  =   8D
	retlw   b'00000000'  ;  >   00
	retlw   b'01100110'  ;  ?   66
	retlw   b'11010110'  ;  @   D6
	retlw   b'01000010'  ;  A   42
	retlw   b'10000100'  ;  B   84
	retlw   b'10100100'  ;  C   A4
	retlw   b'10000011'  ;  D   83
	retlw   b'00000001'  ;  E   01
	retlw   b'00100100'  ;  F   24
	retlw   b'11000011'  ;  G   C3
	retlw   b'00000100'  ;  H   04
	retlw   b'00000010'  ;  I   02
	retlw   b'01110100'  ;  J   74
	retlw   b'10100011'  ;  K   A3
	retlw   b'01000100'  ;  L   44
	retlw   b'11000010'  ;  M   C2
	retlw   b'10000010'  ;  N   82
	retlw   b'11100011'  ;  O   E3
	retlw   b'01100100'  ;  P   64
	retlw   b'11010100'  ;  Q   D4
	retlw   b'01000011'  ;  R   43
	retlw   b'00000011'  ;  S   03
	retlw   b'10000001'  ;  T   81
	retlw   b'00100011'  ;  U   23
	retlw   b'00010100'  ;  V   14
	retlw   b'01100011'  ;  W   63
	retlw   b'10010100'  ;  X   94
	retlw   b'10110100'  ;  Y   B4
	retlw   b'11000100'  ;  Z   C4


;================================
 
;--------------------------------
    cblock 0x20
        Temp
        DelCount, DelCount2
        IntCountLo,  IntCountHi
	TimCountLo, TimCountHi
	FreqCache
        Counter
        SymbCount       ;Counts 0 to 126 fior each WSPR symbol
        WSPRData         ;Used in getting the freq data from EE
        PosData         ;Position of the 2 data bits in Byte from EE
        SoftCount       ;Used in amplitude ramping
        Amplitude
        Flags
        Wreg
        SaveSts
        Seconds, Minutes
	BCDHi, BCDLo
        ICount
        CommaCount
        CWChar, CWMsgCnt, CWCharCnt
        RxChar
        TonePtr
        SerCounter      
        AddrHold
        MsgCount
	D0, D1, D2, D3	;Frequency Data sent on serial interface
        Stack0, Stack1, Stack2, Stack3, Stack4, Stack5, Stack6, Stack7, 
	Stack8, Stack9, StackA, StackB, StackC, StackD, StackE, StackF
    endc   
;==========================================
		end
